//
// Copyright (c) Erinors 2006-2007
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package com.erinors.tapestry.tapdoc.maven;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import org.apache.commons.vfs.FileObject;
import org.apache.commons.vfs.FileSystemManager;
import org.apache.commons.vfs.VFS;
import org.apache.hivemind.ClassResolver;
import org.apache.hivemind.impl.DefaultClassResolver;
import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.DependencyResolutionRequiredException;
import org.apache.maven.artifact.factory.ArtifactFactory;
import org.apache.maven.artifact.handler.ArtifactHandler;
import org.apache.maven.artifact.metadata.ArtifactMetadataSource;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.artifact.resolver.AbstractArtifactResolutionException;
import org.apache.maven.artifact.resolver.ArtifactResolutionResult;
import org.apache.maven.artifact.resolver.ArtifactResolver;
import org.apache.maven.doxia.siterenderer.Renderer;
import org.apache.maven.doxia.siterenderer.RendererException;
import org.apache.maven.doxia.siterenderer.sink.SiteRendererSink;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.artifact.InvalidDependencyVersionException;
import org.apache.maven.reporting.AbstractMavenReport;
import org.apache.maven.reporting.MavenReportException;

import com.erinors.tapestry.tapdoc.util.TapdocPrefs;
import com.erinors.tapestry.tapdoc.util.TapdocUtils;
import org.codehaus.doxia.sink.Sink;

/**
 * @author Norbert S??ndor
 * @goal tapdoc
 * @phase pre-site
 */
public class TapdocReport extends AbstractMavenReport
{
    /**
     * @parameter expression="${plugin.artifacts}"
     */
    private List<Artifact> pluginArtifacts;

    /**
     * The projects in the reactor for aggregation report.
     * 
     * @parameter expression="${reactorProjects}"
     * @readonly
     */
    private List<MavenProject> reactorProjects;

    /**
     * Used for resolving artifacts
     * 
     * @component
     */
    private ArtifactResolver resolver;

    /**
     * Factory for creating artifact objects
     * 
     * @component
     */
    private ArtifactFactory factory;

    /**
     * The local repository where the artifacts are located
     * 
     * @parameter expression="${localRepository}"
     */
    private ArtifactRepository localRepository;

    /**
     * Used to resolve artifacts of aggregated modules
     * 
     * @component
     */
    private ArtifactMetadataSource artifactMetadataSource;

    /**
     * Whether to build an aggregated report at the root, or build individual reports.
     * 
     * @parameter expression="${aggregate}" default-value="false"
     */
    protected boolean aggregate;

    public void execute() throws MojoExecutionException
    {
        super.execute();

        try
        {
            SiteRendererSink sink = siteRenderer.createSink(getReportOutputDirectory(), getOutputName() + ".html");
            generate(sink, Locale.getDefault());
        }
        catch (RendererException e)
        {
            throw new MojoExecutionException("An error has occurred in " + getName(Locale.ENGLISH)
                    + " report generation:" + e.getMessage(), e);
        }
        catch (IOException e)
        {
            throw new MojoExecutionException("An error has occurred in " + getName(Locale.ENGLISH)
                    + " report generation:" + e.getMessage(), e);
        }
        catch (MavenReportException e)
        {
            throw new MojoExecutionException("An error has occurred in " + getName(Locale.ENGLISH)
                    + " report generation:" + e.getMessage(), e);
        }
    }

    public void generate(Sink sink, Locale locale) throws MavenReportException
    {
        outputDirectory = getReportOutputDirectory();

        executeReport(locale);
    }

    protected void executeReport(Locale defaultLocale) throws MavenReportException
    {
        if (aggregate && !project.isExecutionRoot())
        {
            return;
        }
        TapdocPrefs prefs = new TapdocPrefs(showErrors, ignoreAbstract);
        prefs.setLibraryLocations(libraries);

        //
        // Run javadoc
        //
        Collection<CharSequence> classpath = getClasspath();
        String docletpath = null;

        for (Artifact pluginArtifact : pluginArtifacts)
        {
            // TODO only add the really necessary artifacts: tapestry, hivemind, tapdoc
            classpath.add(pluginArtifact.getFile().getAbsolutePath());

            if (pluginArtifact.getGroupId().equals("net.sf.tacos")
                    && pluginArtifact.getArtifactId().equals("tapdoc"))
            {
                docletpath = pluginArtifact.getFile().getAbsolutePath();
            }
        }
        assert docletpath != null;        

        Set<CharSequence> sourcePaths = new HashSet<CharSequence>(project.getCompileSourceRoots());
        if (aggregate && project.isExecutionRoot())
        {
            for (MavenProject subProject : reactorProjects)
            {
                List<CharSequence> sourceRoots = subProject.getCompileSourceRoots();

                if (subProject.getExecutionProject() != null)
                {
                    sourceRoots.addAll(subProject.getExecutionProject().getCompileSourceRoots());
                }

                ArtifactHandler artifactHandler = subProject.getArtifact().getArtifactHandler();
                if ("java".equals(artifactHandler.getLanguage()))
                {
                    sourcePaths.addAll(sourceRoots);
                }
            }
        }
        
        if (packages == null) 
        {
            StringBuffer sb = new StringBuffer();
            for (String lib : libraries)
            {
                lib = lib.trim();
                if (lib.startsWith("/"))
                    lib = lib.substring(1);
                int last = lib.lastIndexOf('/');
                if (last>=0)
                {
                    if (sb.length()>0)
                        sb.append(":");
                    sb.append(lib.substring(0, last).replace('/', '.'));
                }
            }
            packages = sb.toString();
        }
        
        JavadocExecutor javadocExecutor = useJavadocExecutable ? 
            new DirectJavadocExecutor() : new CommandlineJavadocExecutor(trimClasspath);
        
        javadocExecutor.runJavadoc(classpath, docletpath, 
                "com.erinors.tapestry.tapdoc.doclet.TapdocDoclet",
                outputDirectory.getAbsolutePath(), sourcePaths,
                packages);

        //
        // Run Tapdoc
        //
        ClassLoader oldContextClassloader = Thread.currentThread().getContextClassLoader();

        try
        {
            URLClassLoader classLoader = createClassloader(getClasspath());
            Thread.currentThread().setContextClassLoader(classLoader);

            ClassResolver resolver = new DefaultClassResolver(classLoader);

            FileSystemManager fsManager = VFS.getManager();
            FileObject outputDirectoryFile = fsManager.resolveFile("file://"
                    + getReportOutputDirectory().getAbsolutePath());
            FileObject generatedOutputDirectoryFile = fsManager.resolveFile("file://"
                    + generatedOutputDirectory.getAbsolutePath());
            TapdocUtils.generate(resolver, outputDirectoryFile, generatedOutputDirectoryFile, null, prefs);
        }
        catch (Exception e)
        {
            throw new RuntimeException(e);
        }
        finally
        {
            Thread.currentThread().setContextClassLoader(oldContextClassloader);
        }
    }

    private URLClassLoader createClassloader(List<String> classpathFiles) throws MalformedURLException,
            DependencyResolutionRequiredException
    {
        Set<URL> urls = new HashSet<URL>();
        for (String filename : classpathFiles)
        {
            urls.add(new File(filename).toURL());
        }

        if (getClass().getClassLoader() instanceof URLClassLoader)
        {
            for (URL url : ((URLClassLoader) getClass().getClassLoader()).getURLs())
            {
                urls.remove(url);
            }
        }

        //System.out.println(urls);
        return new URLClassLoader(urls.toArray(new URL[] {}), getClass().getClassLoader());
    }

    /**
     * Library specification locations of the libraries to be documented, separated by comma. Eg.
     * &quot;/org/apache/tapestry/Framework.library, /org/apache/tapestry/wml/WML.library&quot;
     * 
     * @parameter
     * @required
     */
    private List<String> libraries;

    /**
     * Directory where reports will go.
     *
     * @parameter expression="${project.build.directory}/tapdocs"
     * @required
     * @readonly
     */
    private File outputDirectory;

    /**
     * Whether to log to the console any runtime errors.
     *
     * @parameter default-value="false"
     */
    protected boolean showErrors;

    /**
     * Ignore abstract classes. By default, all abstract classes
     * found are considered tapestry components.
     *
     * @parameter default-value="false"
     */
    protected boolean ignoreAbstract;
    
    /**
     * The custom doclet that tapdoc must apply to the source code has
     * traditionally used 'java ... com.sun.tools.javadoc.Main ...'.
     * However, there's now an option to execute 'javadoc ...' instead,
     * so, if the default fails, try using this option.
     *
     * @parameter default-value="false"
     */
    protected boolean useJavadocExecutable;    
    
    /**
     * In several projects, the discovered classpath can grow very large,
     * causing the java/javadoc execution to fail due to OS command limits.
     * With this setting on (the default), tapdoc will construct a classpath
     * that only includes tapestry, hivemind and tacos jars.
     *
     * @parameter default-value="true"
     */
    protected boolean trimClasspath;    
    
    /**
     * Packages to search for components (separated by ':').
     * If not specified, it's derived from the libraries parameter.
     *
     * @parameter
     */
    protected String packages;    
    
    protected String getOutputDirectory()
    {
        return outputDirectory.getAbsolutePath();
    }

    /**
     * Directory where reports will go.
     * 
     * @parameter expression="${project.build.directory}/generated-site/xdoc/tapdocs"
     * @required
     * @readonly
     */
    private File generatedOutputDirectory;

    public File getGeneratedOutputDirectory()
    {
        return generatedOutputDirectory;
    }

    /**
     * @parameter expression="${project}"
     * @required
     * @readonly
     */
    private MavenProject project;

    protected MavenProject getProject()
    {
        return project;
    }

    /**
     * @component
     */
    private Renderer siteRenderer;

    protected Renderer getSiteRenderer()
    {
        return siteRenderer;
    }

    /**
     * Specifies the destination directory where Tapdoc saves the generated HTML files.
     * 
     * @parameter expression="${project.reporting.outputDirectory}/tapdoc"
     * @required
     */
    private File reportOutputDirectory;

    public File getReportOutputDirectory()
    {
        if (reportOutputDirectory == null)
        {
            return outputDirectory;
        }

        return reportOutputDirectory;
    }

    public void setReportOutputDirectory(File reportOutputDirectory)
    {
        if ((reportOutputDirectory != null) && (!reportOutputDirectory.getAbsolutePath().endsWith(destDir)))
        {
            this.reportOutputDirectory = new File(reportOutputDirectory, destDir);
        }
        else
        {
            this.reportOutputDirectory = reportOutputDirectory;
        }
    }

    /**
     * The name of the destination directory.
     * 
     * @parameter expression="${destDir}" default-value="tapdocs"
     */
    private String destDir;

    /**
     * The description of the tapdoc report.
     * 
     * @parameter expression="${description}" default-value="Tapestry component reference."
     */
    private String description;

    public String getDescription(Locale locale)
    {
        return description;
    }

    /**
     * The name of the tapdoc report.
     * 
     * @parameter expression="${name}" default-value="TapDocs"
     */
    private String name;

    public String getName(Locale locale)
    {
        return name;
    }

    public String getOutputName()
    {
        return destDir + "/index";
    }

    public boolean isExternalReport()
    {
        return true;
    }

    public String getCategoryName()
    {
        return CATEGORY_PROJECT_REPORTS;
    }

    private List getClasspath() throws MavenReportException
    {
        List classpathElements = new ArrayList();
        Map compileArtifactMap = new HashMap();

        classpathElements.add(project.getBuild().getOutputDirectory());

        populateCompileArtifactMap(compileArtifactMap, project.getCompileArtifacts());

        if (aggregate && project.isExecutionRoot())
        {
            try
            {
                for (Iterator i = reactorProjects.iterator(); i.hasNext();)
                {
                    MavenProject subProject = (MavenProject) i.next();
                    if (subProject != project)
                    {
                        classpathElements.add(subProject.getBuild().getOutputDirectory());
                        Set dependencyArtifacts = subProject.createArtifacts(factory, null, null);
                        if (!dependencyArtifacts.isEmpty())
                        {
                            ArtifactResolutionResult result = resolver.resolveTransitively(dependencyArtifacts,
                                    subProject.getArtifact(), subProject.getRemoteArtifactRepositories(),
                                    localRepository, artifactMetadataSource);
                            populateCompileArtifactMap(compileArtifactMap, getCompileArtifacts(result.getArtifacts()));
                        }
                    }
                }
            }
            catch (AbstractArtifactResolutionException e)
            {
                throw new MavenReportException(e.getMessage(), e);
            }
            catch (InvalidDependencyVersionException e)
            {
                throw new MavenReportException(e.getMessage(), e);
            }
        }

        classpathElements.addAll(compileArtifactMap.values());

        return classpathElements;
    }

    private void populateCompileArtifactMap(Map compileArtifactMap, List artifactList) throws MavenReportException
    {
        if (artifactList != null)
        {
            for (Iterator i = artifactList.iterator(); i.hasNext();)
            {
                Artifact a = (Artifact) i.next();

                File file = a.getFile();

                if (file == null)
                {
                    throw new MavenReportException(
                            "Error in plugin descriptor - compile dependencies were not resolved");
                }
                compileArtifactMap.put(a.getDependencyConflictId(), file.getAbsolutePath());
            }
        }
    }

    private static List getCompileArtifacts(Set artifacts)
    {
        List list = new ArrayList(artifacts.size());

        for (Iterator i = artifacts.iterator(); i.hasNext();)
        {
            Artifact a = (Artifact) i.next();

            // TODO: classpath check doesn't belong here - that's the other method
            if (a.getArtifactHandler().isAddedToClasspath())
            {
                // TODO: let the scope handler deal with this
                if (Artifact.SCOPE_COMPILE.equals(a.getScope()) || Artifact.SCOPE_PROVIDED.equals(a.getScope())
                        || Artifact.SCOPE_SYSTEM.equals(a.getScope()))
                {
                    list.add(a);
                }
            }
        }
        return list;
    }
}
// FIXME annotation-based documentation generation does not work correctly for modules of modules, like
// tapdoc-project/tapdoc-demo/tapdoc-demo-library
